#!/usr/bin/env python3

# Tests for ViZDoom enums and related methods.
# This test can be run as Python script or via PyTest

import vizdoom as vzd


def test_game_args():
    print("Testing setting custom game arguments...")

    game = vzd.DoomGame()
    game.set_window_visible(False)

    args1 = "-deathmatch +timelimit 1 +sv_spawnfarthest 1"
    args2 = "+name ViZDoom +colorset 0"
    args_all = args1 + " " + args2

    game.set_game_args(args_all)
    assert (
        game.get_game_args() == args_all
    ), f"Game args not set correctly. Expected: {args_all}, Got: {game.get_game_args()}"

    game.clear_game_args()
    game.add_game_args(args1)
    game.add_game_args(args2)
    assert (
        game.get_game_args() == args_all
    ), f"Game args not set correctly. Expected: {args_all}, Got: {game.get_game_args()}"


if __name__ == "__main__":
    test_game_args()



#!/usr/bin/env python3

# Tests for make_action method.
# This test can be run as Python script or via PyTest

import numpy as np

import vizdoom as vzd


def _init_game():
    game = vzd.DoomGame()
    game.set_window_visible(False)
    game.set_available_buttons(
        [vzd.Button.MOVE_LEFT, vzd.Button.MOVE_RIGHT, vzd.Button.ATTACK]
    )
    game.set_episode_start_time(35)
    game.init()

    return game


def _test_exception(func, error, msg):
    try:
        func()
    except error:
        pass
    else:
        assert False, msg


def _compare_actions(action_a, action_b):
    if isinstance(action_a, np.ndarray):
        assert np.equal(action_a, action_b).all()
    else:
        assert action_a == action_b


def _test_make_action_input(type_name, type_args={}):
    def _make_func(game, action, tics=1):
        return game.make_action(action, tics)

    def _set_advance_func(game, action, tics=1):
        game.set_action(action)
        game.advance_action(tics, True)
        return game.get_last_reward()

    __test_make_action_input(_make_func, type_name, type_args=type_args)
    __test_make_action_input(_set_advance_func, type_name, type_args=type_args)


def __test_make_action_input(
    make_action_func,
    type_name,
    type_args={},
):
    print(
        f"Testing make_action with {type_name.__name__}([], {type_args}) type as input ..."
    )

    # Prepare game
    game = _init_game()

    prev_pos_y = game.get_game_variable(vzd.GameVariable.POSITION_Y)
    prev_ammo = game.get_game_variable(vzd.GameVariable.AMMO2)

    # make_action() with correct arguments
    next_action = type_name([1, 0, 1], **type_args)
    make_action_func(game, next_action, 8)
    assert prev_pos_y < game.get_game_variable(vzd.GameVariable.POSITION_Y)
    prev_pos_y = game.get_game_variable(vzd.GameVariable.POSITION_Y)
    assert prev_ammo > game.get_game_variable(vzd.GameVariable.AMMO2)
    prev_ammo = game.get_game_variable(vzd.GameVariable.AMMO2)
    _compare_actions(next_action, type_name(game.get_last_action(), **type_args))

    # make_action() without skipping frames
    make_action_func(game, next_action)

    # make_action_func() with negative frames and other types
    error_msg = "make_action() should raise TypeError when called with negative frames or type other than unsigned int"
    _test_exception(
        lambda: make_action_func(game, next_action, -10), TypeError, error_msg
    )
    _test_exception(
        lambda: make_action_func(game, next_action, "10"), TypeError, error_msg
    )

    # make_action() with too short action
    next_action = type_name([1, 0], **type_args)
    make_action_func(game, next_action, 8)
    assert prev_pos_y < game.get_game_variable(vzd.GameVariable.POSITION_Y)
    prev_pos_y = game.get_game_variable(vzd.GameVariable.POSITION_Y)
    _compare_actions(next_action, type_name(game.get_last_action()[:2], **type_args))

    # make_action() with too long action
    next_action = type_name([1, 0, 1, 0], **type_args)
    make_action_func(game, next_action, 8)
    assert prev_pos_y < game.get_game_variable(vzd.GameVariable.POSITION_Y)
    prev_pos_y = game.get_game_variable(vzd.GameVariable.POSITION_Y)
    assert prev_ammo > game.get_game_variable(vzd.GameVariable.AMMO2)
    prev_ammo = game.get_game_variable(vzd.GameVariable.AMMO2)
    _compare_actions(next_action[:3], type_name(game.get_last_action(), **type_args))

    # make_action() with values other than 0 and 1
    next_action = type_name([0, 5, -5], **type_args)
    make_action_func(game, next_action, 16)
    assert prev_pos_y > game.get_game_variable(vzd.GameVariable.POSITION_Y)
    assert prev_ammo > game.get_game_variable(vzd.GameVariable.AMMO2)
    _compare_actions(next_action[:3], type_name(game.get_last_action(), **type_args))


def test_make_action_list():
    def typed_list(data, dtype=int):
        return [dtype(d) for d in data]

    _test_make_action_input(typed_list, {"dtype": int})
    _test_make_action_input(typed_list, {"dtype": float})
    _test_make_action_input(typed_list, {"dtype": bool})


def test_make_action_numpy():
    _test_make_action_input(np.array, {"dtype": int})
    _test_make_action_input(np.array, {"dtype": float})
    _test_make_action_input(np.array, {"dtype": bool})

    _test_make_action_input(np.array, {"dtype": np.float16})
    _test_make_action_input(np.array, {"dtype": np.float32})
    _test_make_action_input(np.array, {"dtype": np.float64})
    _test_make_action_input(np.array, {"dtype": np.int16})
    _test_make_action_input(np.array, {"dtype": np.int32})
    _test_make_action_input(np.array, {"dtype": np.int64})


def test_make_action_tuple():
    def typed_tuple(data, dtype=int):
        return tuple([dtype(d) for d in data])

    _test_make_action_input(typed_tuple, {"dtype": int})
    _test_make_action_input(typed_tuple, {"dtype": float})
    _test_make_action_input(typed_tuple, {"dtype": bool})


def test_make_action_mixed_list():
    def mixed_typed_list(data, dtypes):
        return [dtypes[i](d) for i, d in enumerate(data)]

    _test_make_action_input(mixed_typed_list, {"dtypes": [int, float, bool, int]})
    _test_make_action_input(mixed_typed_list, {"dtypes": [float, bool, int, float]})
    _test_make_action_input(mixed_typed_list, {"dtypes": [bool, int, float, bool]})


def test_keyword_arguments():
    game = _init_game()

    game.make_action(action=[1, 1, 1], tics=10)
    game.set_action(action=[1, 0, 1])
    game.advance_action(update_state=False)
    game.advance_action(tics=10, update_state=True)
    game.advance_action(tics=1)


def test_advance_action():
    game = _init_game()

    # advance_action() without set_action()
    game.advance_action()

    # advance_action() with update_state argument set to False
    state = game.get_state()
    assert state is not None
    prev_tic = state.tic
    tics_to_advance = 10
    game.advance_action(tics_to_advance, update_state=False)
    state = game.get_state()
    assert state is not None
    assert state.tic == prev_tic
    game.advance_action()
    state = game.get_state()
    assert state is not None
    assert state.tic == prev_tic + tics_to_advance + 1


if __name__ == "__main__":
    test_make_action_list()
    test_make_action_numpy()
    test_make_action_tuple()
    test_make_action_mixed_list()
    test_keyword_arguments()
    test_advance_action()



import pytest

import vizdoom as vzd


def test_vizdoom_is_not_running_exception():
    print("Testing ViZDoomIsNotRunningException...")

    game = vzd.DoomGame()
    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.advance_action()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.make_action([0, 1, 0])

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_state()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_server_state()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_last_action()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_last_reward()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_total_reward()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.new_episode()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.replay_episode("demo.replay")

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.is_new_episode()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.is_episode_finished()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.is_episode_timeout_reached()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.is_player_dead()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.respawn_player()

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_button(vzd.Button.ATTACK)

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.get_game_variable(vzd.GameVariable.HEALTH)

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.send_game_command("give ammo")

    with pytest.raises(vzd.ViZDoomIsNotRunningException):
        game.save("non_existent_file.save")


def test_file_does_not_exist_exception():
    print("Testing FileDoesNotExistException...")

    game = vzd.DoomGame()
    with pytest.raises(vzd.FileDoesNotExistException):
        game.load_config("non_existent_file.cfg")

    game = vzd.DoomGame()
    game.set_doom_scenario_path("non_existent_file.wad")
    with pytest.raises(vzd.FileDoesNotExistException):
        game.init()

    game = vzd.DoomGame()
    game.set_doom_game_path("non_existent_file.wad")
    with pytest.raises(vzd.FileDoesNotExistException):
        game.init()


def test_vizdoom_no_sound_exception():
    print("Testing ViZDoomNoOpenALSoundException...")

    # Testing no sound device available with audio buffer enabled (should raise)
    game = vzd.DoomGame()
    game.set_audio_buffer_enabled(True)
    game.set_window_visible(False)
    # game.add_game_args("+snd_backend null")
    # with pytest.raises(vzd.ViZDoomNoOpenALSoundException):
    #     game.init()

    # Testing no sound device available with audio buffer disabled (should not raise)
    game.close()
    game.set_audio_buffer_enabled(False)
    game.init()
    state = game.get_state()
    assert state is not None
    assert state.audio_buffer is None
    game.close()

    # Testing with sound device available and audio buffer enabled (should not raise)
    game.set_audio_buffer_enabled(True)
    game.init()
    state = game.get_state()
    assert state is not None
    assert state.audio_buffer is not None
    game.close()


if __name__ == "__main__":
    test_vizdoom_is_not_running_exception()
    test_file_does_not_exist_exception()
    test_vizdoom_no_sound_exception()



#!/usr/bin/env python3

# Tests ViZDoom load_config method and all the config files from the scenario directory.
# This test can be run as Python script or via PyTest

import os

import vizdoom as vzd


config_values = {
    # "ammo_reward": 0.1,
    "audio_buffer_enabled": True,
    "audio_buffer_size": 8,
    "audio_sampling_rate": vzd.SamplingRate.SR_44100,
    "automap_buffer_enabled": True,
    "automap_mode": vzd.AutomapMode.OBJECTS,
    "automap_render_textures": False,
    "automap_render_objects_as_sprites": True,
    "automap_rotate": True,
    "available_buttons": [vzd.Button.MOVE_FORWARD, vzd.Button.TURN_LEFT],
    "available_game_variables": [vzd.GameVariable.AMMO2, vzd.GameVariable.HEALTH],
    "console_enabled": True,
    "damage_made_reward": 0.1,
    "damage_taken_penalty": 0.1,
    "damage_taken_reward": -0.1,
    "death_penalty": 100,
    "death_reward": -100,
    "depth_buffer_enabled": True,
    "doom_config_path": "test/path/config.cfg"
    if os.name != "nt"
    else "test\\path\\config.cfg",
    "doom_game_path": "test/path/doom.wad"
    if os.name != "nt"
    else "test\\path\\doom.wad",
    "doom_map": "map02",
    "doom_scenario_path": "test/path/scenario.cfg"
    if os.name != "nt"
    else "test\\path\\scenario.cfg",
    "doom_skill": 4,
    "episode_start_time": 2,
    "episode_timeout": 10,
    "frag_reward": 10,
    "game_args": "-fast -respawn",
    "health_reward": 0.1,
    "hit_reward": 0.5,
    "hit_taken_penalty": 0.5,
    "hit_taken_reward": -0.5,
    "item_reward": 1,
    "kill_reward": 5,
    "labels_buffer_enabled": True,
    "living_reward": 2,
    "mode": vzd.Mode.PLAYER,
    "notifications_buffer_enabled": True,
    "notifications_buffer_size": 8,
    "objects_info_enabled": True,
    "render_all_frames": True,
    "render_corpses": False,
    "render_crosshair": True,
    "render_decals": False,
    "render_effects_sprites": False,
    "render_hud": True,
    "render_messages": True,
    "render_minimal_hud": True,
    "render_particles": False,
    "render_screen_flashes": False,
    "render_weapon": False,
    "screen_format": vzd.ScreenFormat.BGR24,
    "screen_resolution": vzd.ScreenResolution.RES_640X480,
    "secret_reward": 50,
    "sectors_info_enabled": True,
    "seed": 1993,
    "sound_enabled": True,
    "ticrate": 70,
    "vizdoom_path": "test/path/vizdoom" if os.name != "nt" else "test\\path\\vizdoom",
    "window_visible": True,
}


def _create_config_string(config_values):
    test_config = ""
    for key, value in config_values.items():
        if isinstance(value, list):
            value = [v.name if hasattr(v, "name") else v for v in value]
            value = "{" + " ".join(value) + "}"
        else:
            if hasattr(value, "name"):
                value = value.name
        test_config += f"{key} = {value}\n"

    return test_config


def _test_if_config_set(game, config_values):
    for key, value in config_values.items():
        # Check if set method exists
        setter = f"set_{key}"
        assert hasattr(
            game, setter
        ), f"Config key {key} does not have a setter method ({setter})"

        if isinstance(value, bool):
            getter = f"is_{key}"
        else:
            getter = f"get_{key}"

        if not hasattr(game, getter):
            print(f"Skipping {key} as there is no getter method ({getter})")
            continue  # Skip if the getter does not exist

        getter_value = getattr(game, getter)()
        print(f"Testing {key}: {getter_value} == {value}")
        assert getter_value == value or str(getter_value) in str(
            value
        ), f"Config value for {key} does not match: expected {value}, got {getattr(game, getter)()}"


def _test_load_config(config_values_dict, remove_underscores):
    print("Testing loading all config keys from a config file ...")

    if remove_underscores:
        _config_values_dict = {
            key.replace("_", ""): value for key, value in config_values_dict.items()
        }
    else:
        _config_values_dict = config_values_dict.copy()

    # Create temporary config file
    with open("test_configs.cfg", "w") as f:
        f.write(_create_config_string(_config_values_dict))

    game = vzd.DoomGame()
    success = game.load_config("test_configs.cfg")
    assert success, "load_config should return True for valid config file"

    _test_if_config_set(game, config_values)

    os.remove("test_configs.cfg")


def _test_set_config(config_values_dict, remove_underscores):
    print("Testing setting all config keys ...")

    if remove_underscores:
        _config_values_dict = {
            key.replace("_", ""): value for key, value in config_values_dict.items()
        }
    else:
        _config_values_dict = config_values_dict.copy()

    # As dictionary
    game = vzd.DoomGame()
    success = game.set_config(_config_values_dict)
    assert success, "set_config should return True for valid config dict"
    _test_if_config_set(game, config_values_dict)

    # Create config string
    game = vzd.DoomGame()
    test_config_str = _create_config_string(_config_values_dict)
    success = game.set_config(test_config_str)
    assert success, "set_config should return True for valid config string"

    _test_if_config_set(game, config_values_dict)


def test_load_config():
    _test_load_config(config_values, False)
    _test_load_config(config_values, True)


def test_set_config():
    _test_set_config(config_values, False)
    _test_set_config(config_values, True)


def test_scenario_configs():
    print("Testing load_config() and default scenarios ...")

    for file in os.listdir(vzd.scenarios_path):
        print(f"Testing for config file: {vzd.scenarios_path}/{file}")
        if file.endswith(".cfg"):
            game = vzd.DoomGame()

            # Both should work
            success = game.load_config(os.path.join(vzd.scenarios_path, file))
            assert (
                success
            ), f"Failed to load config file: {os.path.join(vzd.scenarios_path, file)}"

            success = game.load_config(file)
            assert success, f"Failed to load config file: {file}"

            w = game.get_screen_width()
            h = game.get_screen_height()
            assert (
                w == 320 and h == 240
            ), f"Config file {file} is using non-default screen resolution: {w}x{h} instead 320x240."


if __name__ == "__main__":
    test_set_config()
    test_load_config()
    test_scenario_configs()



#!/usr/bin/env python3

# Tests for get_state() method and returned State object.
# This test can be run as Python script or via PyTest.

import os
import pickle
import random
from itertools import product

import numpy as np
import psutil

import vizdoom as vzd


def _test_get_state(
    num_iterations=10,
    num_states=20,
    mem_eta_mb=0,
    depth_buffer=False,
    labels_buffer=False,
    automap_buffer=False,
    objects_info=False,
    sectors_info=False,
    audio_buffer=False,
    notifications_buffer=False,
    seed=1993,
):
    print("Testing get_state() ...")

    random.seed(seed)

    buttons = [
        vzd.Button.MOVE_FORWARD,
        vzd.Button.MOVE_BACKWARD,
        vzd.Button.MOVE_LEFT,
        vzd.Button.MOVE_RIGHT,
        vzd.Button.TURN_LEFT,
        vzd.Button.TURN_RIGHT,
        vzd.Button.ATTACK,
        vzd.Button.USE,
    ]
    actions = [list(i) for i in product([0, 1], repeat=len(buttons))]

    game = vzd.DoomGame()
    game.set_window_visible(False)
    game.set_episode_timeout(num_states)
    game.set_available_buttons(buttons)

    game.set_depth_buffer_enabled(depth_buffer)
    game.set_labels_buffer_enabled(labels_buffer)
    game.set_automap_buffer_enabled(automap_buffer)
    game.set_objects_info_enabled(objects_info)
    game.set_sectors_info_enabled(sectors_info)
    game.set_audio_buffer_enabled(audio_buffer)
    game.set_notifications_buffer_enabled(notifications_buffer)

    buffers = ["screen_buffer"]
    if depth_buffer:
        buffers.append("depth_buffer")
    if labels_buffer:
        buffers.append("labels_buffer")
    if automap_buffer:
        buffers.append("automap_buffer")
    if audio_buffer:
        buffers.append("audio_buffer")
        # This fixes "BiquadFilter_setParams: Assertion `gain > 0.00001f' failed" issue
        # or "no audio in buffer" issue caused by a bug in OpenAL version 1.19.
        game.add_game_args("+snd_efx 0")

    game.init()

    prev_mem = 0
    prev_len = 0
    for i in range(num_iterations):

        states = []
        buffers_copies = []

        game.new_episode()
        for _ in range(num_states):
            if game.is_episode_finished():
                game.new_episode()

            state = game.get_state()

            states.append(state)
            copies = {}
            for b in buffers:
                copies[b] = np.copy(getattr(state, b))

            buffers_copies.append(copies)
            game.make_action(random.choice(actions), 4)

        assert len(states) == num_states
        assert len(buffers_copies) == num_states

        # Compare states with their copies - confirms that states don't mutate.
        # Check min and max values of buffers - confirms that buffers are not empty.
        min_vals = {b: np.inf for b in buffers}
        max_vals = {b: -np.inf for b in buffers}
        for s, bs_copy in zip(states, buffers_copies):
            for b in buffers:
                assert np.array_equal(
                    getattr(s, b), bs_copy[b]
                ), f"Buffer {b} is not equal with its copy"
                min_vals[b] = min(min_vals[b], np.min(bs_copy[b]))
                max_vals[b] = max(max_vals[b], np.max(bs_copy[b]))

        for b in buffers:
            assert (
                min_vals[b] != max_vals[b]
            ), f"Buffer {b} min: {min_vals[b]}, max: {max_vals[b]} are equal, buffer is empty"

        # Save and load states via pickle - confirms that states and all sub-objects (labels, lines, objects) are picklable.
        with open("tmp_states.pkl", "wb") as f:
            pickle.dump(states, f)

        with open("tmp_states.pkl", "rb") as f:
            pickled_states = pickle.load(f)

        # Compare loaded states with their copies - to confirm that pickling doesn't mutate states.
        for s, s_copy in zip(states, pickled_states):
            assert pickle.dumps(s) == pickle.dumps(
                s_copy
            ), "Pickled state is not equal with its original object after save and load"

        del pickled_states
        os.remove("tmp_states.pkl")

        # Check memory for leaks
        process = psutil.Process(os.getpid())
        mem = process.memory_info().rss / 1024 / 1024

        if (i + 1) % 10 == 0:
            print(
                f"Memory, with {len(states)} states saved, after episode {i + 1} / {num_iterations}: {mem} MB, expected ~{prev_mem} +/- {mem_eta_mb} MB"
            )

        if mem_eta_mb > 0:
            if prev_len < len(states):
                prev_mem = mem
                prev_len = len(states)
            elif prev_len == len(states):
                assert (
                    abs(prev_mem - mem) < mem_eta_mb
                ), f"Memory leak detected: with {len(states)} states saved, after episode {i + 1} / {num_iterations}: {mem} MB used, expected ~{prev_mem} +/- {mem_eta_mb} MB"


def test_get_state(num_iterations=10, num_states=20):
    _test_get_state(num_iterations=num_iterations, num_states=num_states, mem_eta_mb=0)
    _test_get_state(
        num_iterations=num_iterations,
        num_states=num_states,
        mem_eta_mb=0,
        depth_buffer=True,
        labels_buffer=True,
        automap_buffer=True,
        objects_info=True,
        sectors_info=True,
        audio_buffer=True,
        notifications_buffer=True,
    )


def _check_state(
    state,
    expect_game_variables=False,
    expect_depth_buffer=False,
    expect_labels_buffer=False,
    expect_automap_buffer=False,
    expect_objects_info=False,
    expect_sectors_info=False,
    expect_audio_buffer=False,
    expect_notifications_buffer=False,
):
    assert state is not None
    assert isinstance(state.tic, int), f"State tic is: {state.tic}, expected int"
    assert isinstance(
        state.screen_buffer, np.ndarray
    ), f"Screen buffer is: {state.screen_buffer}, expected np.ndarray"

    if expect_game_variables:
        assert isinstance(
            state.game_variables, np.ndarray
        ), f"Game variables are: {state.game_variables}, expected np.ndarray"
    else:
        assert (
            state.game_variables is None
        ), f"Game variables are: {state.game_variables}, expected None"

    if expect_depth_buffer:
        assert isinstance(
            state.depth_buffer, np.ndarray
        ), f"Depth buffer is: {state.depth_buffer}, expected np.ndarray"
    else:
        assert (
            state.depth_buffer is None
        ), f"Depth buffer is: {state.depth_buffer}, expected None"

    if expect_labels_buffer:
        assert isinstance(
            state.labels_buffer, np.ndarray
        ), f"Labels buffer is: {state.labels_buffer}, expected np.ndarray"
    else:
        assert (
            state.labels_buffer is None
        ), f"Labels buffer is: {state.labels_buffer}, expected None"

    if expect_automap_buffer:
        assert isinstance(
            state.automap_buffer, np.ndarray
        ), f"Automap buffer is: {state.automap_buffer}, expected np.ndarray"
        assert state.screen_buffer.shape == state.automap_buffer.shape, (
            f"Screen buffer shape is: {state.screen_buffer.shape}, "
            f"automap buffer shape is: {state.automap_buffer.shape}, expected equal",
        )
    else:
        assert (
            state.automap_buffer is None
        ), f"Automap buffer is: {state.automap_buffer}, expected None"

    if expect_objects_info:
        assert isinstance(
            state.objects, list
        ), f"Objects are: {state.objects}, expected list"
    else:
        assert state.objects is None, f"Objects are: {state.objects}, expected None"

    if expect_sectors_info:
        assert isinstance(
            state.sectors, list
        ), f"Sectors are: {state.sectors}, expected list"
    else:
        assert state.sectors is None, f"Sectors are: {state.sectors}, expected None"

    if expect_audio_buffer:
        assert isinstance(
            state.audio_buffer, np.ndarray
        ), f"Audio buffer is: {state.audio_buffer}, expected np.ndarray"
    else:
        assert (
            state.audio_buffer is None
        ), f"Audio buffer is: {state.audio_buffer}, expected None"

    if expect_notifications_buffer:
        assert isinstance(
            state.notifications_buffer, str
        ), f"Notifications buffer is: {state.notifications_buffer}, expected str"
    else:
        assert (
            state.notifications_buffer is None
        ), f"Notifications buffer is: {state.notifications_buffer}, expected None"


def test_buffer_sizes():
    game = vzd.DoomGame()
    game.set_window_visible(False)

    game.set_depth_buffer_enabled(True)
    game.set_labels_buffer_enabled(True)
    game.set_automap_buffer_enabled(True)
    game.set_audio_buffer_enabled(True)
    game.set_notifications_buffer_enabled(True)
    game.set_screen_resolution(vzd.ScreenResolution.RES_640X480)
    game.set_screen_format(vzd.ScreenFormat.CRCGCB)
    game.set_audio_sampling_rate(vzd.SamplingRate.SR_44100)
    game.set_audio_buffer_size(8)
    game.set_notifications_buffer_size(8)

    screen_width = game.get_screen_width()
    screen_height = game.get_screen_height()
    assert screen_width == 640
    assert screen_height == 480

    audio_buf_size = game.get_audio_buffer_size()
    notifications_buf_size = game.get_notifications_buffer_size()
    assert audio_buf_size == 8
    assert notifications_buf_size == 8

    game.init()
    state = game.get_state()
    _check_state(
        state,
        expect_depth_buffer=True,
        expect_labels_buffer=True,
        expect_automap_buffer=True,
        expect_audio_buffer=True,
        expect_notifications_buffer=True,
    )
    assert state.screen_buffer.shape == (3, screen_height, screen_width)  # type: ignore
    assert state.depth_buffer.shape == (screen_height, screen_width)  # type: ignore
    assert state.labels_buffer.shape == (screen_height, screen_width)  # type: ignore
    assert state.automap_buffer.shape == (3, screen_height, screen_width)  # type: ignore
    assert state.audio_buffer.shape == (44100 // 35 * audio_buf_size, 2)  # type: ignore
    assert isinstance(state.notifications_buffer, str)


def test_if_none():
    game = vzd.DoomGame()
    game.set_window_visible(False)
    game.init()

    state = game.get_state()
    _check_state(state)


def test_types():
    game = vzd.DoomGame()
    game.set_window_visible(False)

    game.add_available_game_variable(vzd.GameVariable.AMMO2)
    game.set_depth_buffer_enabled(True)
    game.set_labels_buffer_enabled(True)
    game.set_automap_buffer_enabled(True)
    game.set_objects_info_enabled(True)
    game.set_sectors_info_enabled(True)
    game.set_audio_buffer_enabled(True)
    game.set_notifications_buffer_enabled(True)

    # This fixes "BiquadFilter_setParams: Assertion `gain > 0.00001f' failed" issue
    # or "no audio in buffer" issue caused by a bug in OpenAL version 1.19.
    game.add_game_args("+snd_efx 0")

    game.init()
    state = game.get_state()

    _check_state(
        state,
        expect_game_variables=True,
        expect_depth_buffer=True,
        expect_labels_buffer=True,
        expect_automap_buffer=True,
        expect_objects_info=True,
        expect_sectors_info=True,
        expect_audio_buffer=True,
        expect_notifications_buffer=True,
    )


def test_modifing_buffers_while_game_is_running():
    game = vzd.DoomGame()
    game.set_window_visible(False)
    game.set_screen_resolution(vzd.ScreenResolution.RES_320X240)
    game.set_audio_buffer_size(4)
    game.set_notifications_buffer_size(4)

    # This fixes "BiquadFilter_setParams: Assertion `gain > 0.00001f' failed" issue
    # or "no audio in buffer" issue caused by a bug in OpenAL version 1.19.
    game.add_game_args("+snd_efx 0")

    screen_width = game.get_screen_width()
    screen_height = game.get_screen_height()
    audio_buf_size = game.get_audio_buffer_size()
    notifications_buf_size = game.get_notifications_buffer_size()

    game.init()

    # Initial state - no buffers enabled
    state = game.get_state()
    _check_state(state)

    game.advance_action(4)

    # Game is running, try to change buffers setting
    game.set_depth_buffer_enabled(True)
    game.set_labels_buffer_enabled(True)
    game.set_automap_buffer_enabled(True)
    game.set_objects_info_enabled(True)
    game.set_sectors_info_enabled(True)
    game.set_audio_buffer_enabled(True)
    game.set_notifications_buffer_enabled(True)
    game.set_screen_resolution(vzd.ScreenResolution.RES_640X480)
    game.set_audio_buffer_size(8)
    game.set_notifications_buffer_size(8)

    game.advance_action(4)

    state = game.get_state()
    _check_state(state)

    assert not game.is_depth_buffer_enabled()
    assert not game.is_labels_buffer_enabled()
    assert not game.is_automap_buffer_enabled()
    assert not game.is_objects_info_enabled()
    assert not game.is_sectors_info_enabled()
    assert not game.is_audio_buffer_enabled()
    assert not game.is_notifications_buffer_enabled()
    assert game.get_screen_width() == screen_width == 320
    assert game.get_screen_height() == screen_height == 240
    assert game.get_audio_buffer_size() == audio_buf_size == 4
    assert game.get_notifications_buffer_size() == notifications_buf_size == 4

    game.close()
    game.init()

    state = game.get_state()
    _check_state(state)

    assert not game.is_depth_buffer_enabled()
    assert not game.is_labels_buffer_enabled()
    assert not game.is_automap_buffer_enabled()
    assert not game.is_objects_info_enabled()
    assert not game.is_sectors_info_enabled()
    assert not game.is_audio_buffer_enabled()
    assert not game.is_notifications_buffer_enabled()
    assert game.get_screen_width() == screen_width == 320
    assert game.get_screen_height() == screen_height == 240
    assert game.get_audio_buffer_size() == audio_buf_size == 4
    assert game.get_notifications_buffer_size() == notifications_buf_size == 4


def test_server_state():
    game = vzd.DoomGame()
    game.set_window_visible(False)
    game.init()

    state = game.get_state()
    server_state = game.get_server_state()

    assert server_state is not None
    assert (
        server_state.tic == state.tic
    ), f"State tic is: {state.tic}, server state tic is: {server_state.tic}, expected equal"
    assert (
        server_state.player_count == 1
    ), f"Server state player count is: {server_state.player_count}, expected 1 for singeplayer game"


# TODO: Add more tests for server state


if __name__ == "__main__":
    test_get_state()
    test_buffer_sizes()
    test_if_none()
    test_types()
    test_modifing_buffers_while_game_is_running()
    test_server_state()



#!/usr/bin/env python3

# Tests for Gymnasium wrapper.
# This test can be run as Python script or via PyTest

import os
import pickle

import gymnasium
import numpy as np
from gymnasium.spaces import Box, Dict, Discrete, MultiBinary, MultiDiscrete, Text
from gymnasium.utils.env_checker import check_env, data_equivalence

from vizdoom import gymnasium_wrapper, install_path  # noqa
from vizdoom.gymnasium_wrapper.base_gymnasium_env import ASCII_CHARS, VizdoomEnv


# Ensure pytest.mark.parametrize decorator works without pytest
try:
    import pytest
except ImportError:

    class MockMark:
        def parametrize(self, *args, **kwargs):
            def decorator(func):
                return func

            return decorator

    class MockSkip(Exception):
        """Raised to mimic pytest.skip behavior."""

    def _mock_skip(reason: str = "") -> None:
        raise MockSkip(reason)

    _mock_skip.Exception = MockSkip

    class MockPytest:
        mark = MockMark()
        skip = staticmethod(_mock_skip)

    pytest = MockPytest()
    del MockMark, MockPytest, _mock_skip


VIZDOOM_ENVS = [
    env
    for env in [env_spec.id for env_spec in gymnasium.envs.registry.values()]  # type: ignore
    if "Vizdoom" in env
]
# Skip environments with higher skills levels for testing purposes as they only differ with a single parameter
for skill in ["-S1-", "-S2-", "-S4-", "-S5-"]:
    VIZDOOM_ENVS = [env for env in VIZDOOM_ENVS if skill not in env]

TEST_ENV_CONFIGS = f"{os.path.dirname(os.path.abspath(__file__))}/env_configs"
BUFFERS = ["screen", "depth", "labels", "automap", "audio", "notifications"]

FP32_ACT_SPACE = dict(
    low=np.finfo(np.float32).min, high=np.finfo(np.float32).max, dtype=np.float32
)
TRI_CHANNEL_SCREEN_OBS_SPACE = Box(0, 255, (240, 320, 3), dtype=np.uint8)
SINGLE_CHANNEL_SCREEN_OBS_SPACE = Box(0, 255, (240, 320, 1), dtype=np.uint8)
AUDIO_OBS_SPACE = Box(
    -32768, 32767, (int(44100 * 1 / 35 * 1), 2), dtype=np.int16
)  # sampling rate = 44100, frame_skip = 1
NOTIFICATIONS_OBS_SPACE = Text(min_length=0, max_length=32768, charset=ASCII_CHARS)

COLOR_SCREEN: dict[str, gymnasium.Space] = {"screen": TRI_CHANNEL_SCREEN_OBS_SPACE}
GREY_SCREEN: dict[str, gymnasium.Space] = {"screen": SINGLE_CHANNEL_SCREEN_OBS_SPACE}
DEPTH_BUFFER: dict[str, gymnasium.Space] = {"depth": SINGLE_CHANNEL_SCREEN_OBS_SPACE}
LABELS_BUFFER: dict[str, gymnasium.Space] = {"labels": SINGLE_CHANNEL_SCREEN_OBS_SPACE}
COLOR_AUTOMAP: dict[str, gymnasium.Space] = {"automap": TRI_CHANNEL_SCREEN_OBS_SPACE}
GRAY_AUTOMAP: dict[str, gymnasium.Space] = {"automap": SINGLE_CHANNEL_SCREEN_OBS_SPACE}
NOTIFICATIONS: dict[str, gymnasium.Space] = {"notifications": NOTIFICATIONS_OBS_SPACE}
AUDIO_BUFFER: dict[str, gymnasium.Space] = {"audio": AUDIO_OBS_SPACE}


def _check_if_main_wad_available(env_name: str, env: gymnasium.Env) -> None:
    """
    Helper function to check if specified main WAD file is available for the given environment.
    """
    main_wad_path = env.unwrapped.game.get_doom_game_path()
    if main_wad_path is not None and main_wad_path != "":
        basename = os.path.basename(main_wad_path)
        if (
            not os.path.exists(main_wad_path)
            and not os.path.exists(basename)
            and not os.path.exists(os.path.join(install_path, basename))
        ):
            pytest.skip(
                f"Main WAD file {main_wad_path} not available for {env_name}, skipping test."
            )


def _run_with_pytest_skip(func, *args, **kwargs):
    skip_exception = getattr(getattr(pytest, "skip", None), "Exception", None)
    try:
        func(*args, **kwargs)
    except BaseException as exc:
        if skip_exception is not None and isinstance(exc, skip_exception):
            print(f"Skipped: {exc}")
            return
        raise


# Testing with different non-default kwargs (since each has a different obs space)
# should give warning forcing RGB24 screen type
@pytest.mark.parametrize("env_name", VIZDOOM_ENVS)
def test_gymnasium_wrapper(env_name: str):
    print(f"Testing Gymnasium wrapper - {env_name}")

    for frame_skip in [1, 4]:
        env = gymnasium.make(env_name, frame_skip=frame_skip)
        _check_if_main_wad_available(env_name, env)

        # Test if env adheres to Gymnasium API
        check_env(env.unwrapped, skip_render_check=True)

        ob_space = env.observation_space
        act_space = env.action_space
        obs, _ = env.reset()
        assert ob_space.contains(obs), f"Reset observation: {obs!r} not in space"

        a = act_space.sample()
        observation, reward, terminated, truncated, _info = env.step(a)
        assert ob_space.contains(
            observation
        ), f"Step observation: {observation!r} not in space"
        assert np.isscalar(reward), f"{reward} is not a scalar for {env}"
        assert isinstance(terminated, bool), f"Expected {terminated} to be a boolean"
        assert isinstance(terminated, bool), f"Expected {terminated} to be a boolean"
        assert isinstance(truncated, bool), f"Expected {truncated} to be a boolean"

        env.close()


# Testing obs on terminal state (terminal state is handled differently)
# should give warning forcing RGB24 screen type
@pytest.mark.parametrize("env_name", VIZDOOM_ENVS)
def test_gymnasium_wrapper_terminal_state(env_name: str):
    print(f"Testing Gymnasium terminal state - {env_name}")

    for frame_skip in [1, 4]:
        env = gymnasium.make(env_name, frame_skip=frame_skip, max_buttons_pressed=0)
        _check_if_main_wad_available(env_name, env)

        obs = env.reset()
        terminated = False
        truncated = False
        done = terminated or truncated
        while not done:
            a = env.action_space.sample()
            (obs, _reward, terminated, truncated, _info) = env.step(a)
            done = terminated or truncated
            if done:
                break

        assert env.observation_space.contains(obs)
        env.close()


def test_gymnasium_wrapper_truncated_state():
    print("Testing Gymnasium wrapper truncated state")
    env = VizdoomEnv(
        config_file=os.path.join(
            TEST_ENV_CONFIGS, "basic_rgb_idla_0_1.cfg"
        ),  # For this config it is impossible to get other terminal state than timeout
        frame_skip=10,  # Using frame_skip=10 to speed up the test
        max_buttons_pressed=0,
        treat_episode_timeout_as_truncation=True,
    )

    obs = env.reset()
    terminated = False
    truncated = False
    done = terminated or truncated
    while not done:
        a = env.action_space.sample()
        obs, _reward, terminated, truncated, _info = env.step(a)
        done = terminated or truncated
        if done:
            assert truncated
            env.close()


# Testing various observation spaces
# Using both screen types `(GRAY8, RGB24)` for various combinations of buffers `(screen|depth|labels|automap)`

OBS_SPACE_PARAMS = [
    ("basic_rgb_i_1_3", Dict(COLOR_SCREEN)),
    ("basic_g8_i_1_0", Dict(GREY_SCREEN)),
    ("basic_g8_i_1_0_notifications", Dict(GREY_SCREEN | NOTIFICATIONS)),
    ("basic_g8_i_1_0_audio", Dict(GREY_SCREEN | AUDIO_BUFFER)),
    (
        "basic_g8_idla_4_2",
        Dict(GREY_SCREEN | DEPTH_BUFFER | LABELS_BUFFER | GRAY_AUTOMAP),
    ),
    ("basic_g8_idl_3_1", Dict(GREY_SCREEN | DEPTH_BUFFER | LABELS_BUFFER)),
    ("basic_rgb_id_2_0", Dict(COLOR_SCREEN | DEPTH_BUFFER)),
    (
        "basic_rgb_idla_0_1",
        Dict(COLOR_SCREEN | DEPTH_BUFFER | LABELS_BUFFER | COLOR_AUTOMAP),
    ),
]


@pytest.mark.parametrize("env_config,obs_space", OBS_SPACE_PARAMS)
def test_gymnasium_wrapper_obs_space(env_config: str, obs_space: Dict):
    env = VizdoomEnv(
        config_file=os.path.join(TEST_ENV_CONFIGS, env_config + ".cfg"),
        frame_skip=1,
        max_buttons_pressed=0,
    )
    _check_if_main_wad_available(env_config, env)
    assert env.observation_space == obs_space, (
        f"Incorrect observation space: {env.observation_space!r}, "
        f"should be: {obs_space!r}"
    )
    obs, _ = env.reset()
    assert env.observation_space.contains(
        obs
    ), f"Step observation: {obs!r} not in space"


def _compare_action_spaces(env: gymnasium.Env, expected_action_space: gymnasium.Space):
    """
    Helper function to compare the action space of the environment with the expected action space.
    """
    assert env.action_space == expected_action_space, (
        f"Incorrect action space: {env.action_space!r}, "
        f"should be: {expected_action_space!r}"
    )
    env.reset()
    # check successful call to step using action_space.sample()
    sample_action = env.action_space.sample()
    env.step(sample_action)


# Testing all possible action space combinations
CONTINUOUS_1 = Box(shape=(1,), **FP32_ACT_SPACE)  # type: ignore
CONTINUOUS_2 = Box(shape=(2,), **FP32_ACT_SPACE)  # type: ignore
CONTINUOUS_3 = Box(shape=(3,), **FP32_ACT_SPACE)  # type: ignore

ACTION_SPACE_PARAMS = [
    (
        "basic_rgb_i_1_3",
        Dict({"binary": MultiBinary(1), "continuous": CONTINUOUS_3}),
        Dict({"binary": MultiDiscrete([2]), "continuous": CONTINUOUS_3}),
        [
            Dict({"binary": Discrete(2), "continuous": CONTINUOUS_3}),
            Dict({"binary": Discrete(2), "continuous": CONTINUOUS_3}),
            Dict({"binary": Discrete(2), "continuous": CONTINUOUS_3}),
        ],
    ),
    (
        "basic_g8_i_1_0",
        MultiBinary(1),
        MultiDiscrete([2]),
        [Discrete(2), Discrete(2), Discrete(2)],
    ),
    (
        "basic_g8_idla_4_2",
        Dict({"binary": MultiBinary(4), "continuous": CONTINUOUS_2}),
        Dict({"binary": MultiDiscrete([2, 2, 2, 2]), "continuous": CONTINUOUS_2}),
        [
            Dict({"binary": Discrete(5), "continuous": CONTINUOUS_2}),
            Dict({"binary": Discrete(11), "continuous": CONTINUOUS_2}),
            Dict({"binary": Discrete(15), "continuous": CONTINUOUS_2}),
        ],
    ),
    (
        "basic_g8_idl_3_1",
        Dict({"binary": MultiBinary(3), "continuous": CONTINUOUS_1}),
        Dict({"binary": MultiDiscrete([2, 2, 2]), "continuous": CONTINUOUS_1}),
        [
            Dict({"binary": Discrete(4), "continuous": CONTINUOUS_1}),
            Dict({"binary": Discrete(7), "continuous": CONTINUOUS_1}),
            Dict({"binary": Discrete(8), "continuous": CONTINUOUS_1}),
        ],
    ),
    (
        "basic_rgb_id_2_0",
        MultiBinary(2),
        MultiDiscrete([2, 2]),
        [Discrete(3), Discrete(4), Discrete(4)],
    ),
    (
        "basic_rgb_idla_0_1",
        CONTINUOUS_1,
        CONTINUOUS_1,
        [CONTINUOUS_1, CONTINUOUS_1, CONTINUOUS_1],
    ),
]


@pytest.mark.parametrize(
    "env_config,multi_binary_action_space,multi_discrete_action_space,discrete_action_spaces",
    ACTION_SPACE_PARAMS,
)
def test_gymnasium_wrapper_action_space(
    env_config: str,
    multi_binary_action_space: gymnasium.Space,
    multi_discrete_action_space: gymnasium.Space,
    discrete_action_spaces: list[gymnasium.Space],
):
    env = VizdoomEnv(
        config_file=os.path.join(TEST_ENV_CONFIGS, env_config + ".cfg"),
        frame_skip=1,
        max_buttons_pressed=0,
        use_multi_binary_action_space=True,
    )
    _compare_action_spaces(env, multi_binary_action_space)

    env = VizdoomEnv(
        config_file=os.path.join(TEST_ENV_CONFIGS, env_config + ".cfg"),
        frame_skip=1,
        max_buttons_pressed=0,
        use_multi_binary_action_space=False,
    )
    _compare_action_spaces(env, multi_discrete_action_space)

    for max_button_pressed, action_space in enumerate(discrete_action_spaces, start=1):
        env = VizdoomEnv(
            config_file=os.path.join(TEST_ENV_CONFIGS, env_config + ".cfg"),
            frame_skip=1,
            max_buttons_pressed=max_button_pressed,
        )
        _compare_action_spaces(env, action_space)


def _compare_envs(
    env1, env2, env1_name="First", env2_name="Second", max_steps=10, seed=1993
):
    """
    Helper function to compare two environments.
    It checks if the initial observations, actions, and subsequent observations,
    rewards, termination, truncation, and info are equivalent.
    """
    # Seed environments
    obs1, _ = env1.reset(seed=seed)
    obs2, _ = env2.reset(seed=seed)

    # Seed action space sampler
    env1.action_space.seed(seed)
    env2.action_space.seed(seed)

    assert data_equivalence(
        obs1, obs2
    ), f"Initial observations incorrect. {env1_name} environment: {obs1}. {env2_name} environment: {obs2}"

    # Compare sequance of random actions and states
    done = False
    steps = 0
    while not done and steps < max_steps:
        a1 = env1.action_space.sample()
        a2 = env2.action_space.sample()
        assert data_equivalence(
            a1, a2
        ), f"Actions incorrect. First environment: {a1}. Second environment: {a2}"

        obs1, rew1, term1, trunc1, info1 = env1.step(a1)
        obs2, rew2, term2, trunc2, info2 = env2.step(a2)

        assert data_equivalence(
            obs1, obs2
        ), f"Incorrect observations: {env1_name} environment: {obs1}. {env2_name} environment: {obs2}"
        assert data_equivalence(
            rew1, rew2
        ), f"Incorrect rewards: {env1_name} environment: {rew1}. {env2_name} environment:{rew2}"
        assert data_equivalence(
            term1, term2
        ), f"Incorrect terms: {env1_name} environment: {term1}. {env2_name} environment: {term2}"
        assert data_equivalence(
            trunc1, trunc2
        ), f"Incorrect truncs: {env1_name} environment: {trunc1}. {env2_name} environment: {trunc2}"
        assert data_equivalence(
            info1, info2
        ), f"Incorrect info: {env1_name} environment: {info1}. {env2_name} environment: {info2}"

        done = term1 or trunc1 or term2 or trunc2
        steps += 1

    env1.close()
    env2.close()


@pytest.mark.parametrize("env_name", VIZDOOM_ENVS)
def test_gymnasium_wrapper_pickle(env_name: str):
    print(f"Testing Gymnasium wrapper pickle - {env_name}")

    env1 = gymnasium.make(env_name)
    _check_if_main_wad_available(env_name, env1)
    env2 = pickle.loads(pickle.dumps(env1))

    _compare_envs(
        env1,
        env2,
        env1_name="Original",
        env2_name="Pickled",
        seed=1993,
    )


@pytest.mark.parametrize("env_name", VIZDOOM_ENVS)
def test_gymnasium_wrapper_seed(env_name: str):
    print(f"Testing Gymnasium wrapper seed - {env_name}")

    env1 = gymnasium.make(env_name)
    _check_if_main_wad_available(env_name, env1)
    env2 = gymnasium.make(env_name)

    _compare_envs(
        env1,
        env2,
        env1_name="First",
        env2_name="Second",
        seed=1993,
    )


if __name__ == "__main__":
    print("Testing Gymnasium wrapper compatibility with gymnasium API")
    for env_name in VIZDOOM_ENVS:
        _run_with_pytest_skip(test_gymnasium_wrapper, env_name)

    print("Testing Gymnasium rollout (checking terminal state)")
    for env_name in VIZDOOM_ENVS:
        _run_with_pytest_skip(test_gymnasium_wrapper_terminal_state, env_name)

    test_gymnasium_wrapper_truncated_state()

    print("Testing Gymnasium wrapper action spaces")
    for (
        env_config,
        multi_binary_action_space,
        multi_discrete_action_space,
        discrete_action_spaces,
    ) in ACTION_SPACE_PARAMS:
        _run_with_pytest_skip(
            test_gymnasium_wrapper_action_space,
            env_config,
            multi_binary_action_space,
            multi_discrete_action_space,
            discrete_action_spaces,
        )

    print("Testing Gymnasium wrapper observation spaces")
    for env_config, obs_space in OBS_SPACE_PARAMS:
        _run_with_pytest_skip(test_gymnasium_wrapper_obs_space, env_config, obs_space)

    print("Testing Gymnasium wrapper pickling (EzPickle).")
    for env_name in VIZDOOM_ENVS:
        _run_with_pytest_skip(test_gymnasium_wrapper_pickle, env_name)

    print("Testing gymnasium wrapper seeding.")
    for env_name in VIZDOOM_ENVS:
        _run_with_pytest_skip(test_gymnasium_wrapper_seed, env_name)



#!/usr/bin/env python3

# Tests for category mapping functionality (semantic segmentation).
# This test can be run as Python script or via PyTest

import glob
import os
import random

import vizdoom


def custom_categories_test(scenario: str = "scenarios/defend_the_line.wad"):
    print(f"Testing custom category mapping on {scenario}...")

    # Create game1 instance
    game = vizdoom.DoomGame()

    # Set up basic configuration
    game.set_doom_scenario_path(scenario)
    game.set_window_visible(False)
    game.set_screen_resolution(vizdoom.ScreenResolution.RES_320X240)
    game.set_labels_buffer_enabled(True)
    game.init()

    # Start a new episode
    random.seed("ViZDoom!")
    game.set_seed(random.randrange(0, 256))
    game.new_episode()

    doom_categories = vizdoom.get_default_categories()
    seen_objects = set()
    seen_categories = set()

    # Move randomly for a while to generate some labels
    action_size = [0] * game.get_available_buttons_size()
    for _ in range(2000):
        game.make_action([random.random() > 0.5 for _ in action_size], 4)

        # Get the state and check labels
        state = game.get_state()
        if state and state.labels:
            labels = sorted(state.labels, key=lambda label: label.object_name)
            for label in labels:
                if label.object_category == "Self":
                    assert (
                        label.object_name == "DoomPlayer"
                    ), f'Assigned "Self" to non-DoomPlayer object: {label.object_name}'
                else:
                    assert (
                        label.object_category in doom_categories
                    ), f'Unknown category "{label.object_category}" assigned to {label.object_name}'
                seen_objects.add(label.object_name)
                seen_categories.add(label.object_category)
        else:
            game.new_episode()

    # Close the game
    game.close()
    print(
        f"Seen objects: {', '.join(sorted(seen_objects))}\nSeen categories: {', '.join(sorted(seen_categories))}\n"
    )
    assert "Unknown" not in seen_categories, '"Unknown" category in default scenarios!'


def test_object_categories():
    scenario_pattern = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "scenarios", "*.wad"
    )
    print(f"Finding scenarios in {scenario_pattern}")
    for scenario in sorted(glob.glob(scenario_pattern)):
        if any(multi_keyword in scenario for multi_keyword in ["multi", "cig"]):
            continue
        custom_categories_test(scenario)
    print("\nTest completed!")


if __name__ == "__main__":
    test_object_categories()



#!/usr/bin/env python3


import re

import vizdoom as vzd


def _get_object_methods(_object):
    object_methods = [
        method_name
        for method_name in dir(_object)
        if callable(getattr(_object, method_name)) and not method_name.startswith("_")
    ]
    return object_methods


def _check_object_docstrings(_object):
    object_methods = _get_object_methods(_object)

    for method in object_methods:
        method_doc = getattr(_object, method).__doc__
        assert method_doc is not None, f"Method {method} has no docstring"

        if method.startswith("set_") and method not in {
            "set_action",
            "set_button_max_value",
            "set_config",
        }:
            search_result = re.search(r"Default value: (.+)", method_doc)
            assert search_result is not None, f"Method {method} has no default value"
            # default_value = search_result.group(1).strip(".").strip("`")
            search_result = re.search(r"Config key: ``(.+)``", method_doc)
            assert search_result is not None, f"Method {method} has no config key"
            config_keys = search_result.group(1).split("/")
            config_keys = [key.strip("`") for key in config_keys if key.strip()]
            if len(config_keys) > 1:
                assert (
                    config_keys[1].replace("_", "").lower() == config_keys[0].lower()
                ), f"Config keys do not match: {config_keys}"
            assert (
                method[4:] in config_keys
            ), f"{method} name does not match its config key: {config_keys}"


def test_docstrings():
    print("Testing all docstrings ...")
    _check_object_docstrings(vzd.DoomGame)


if __name__ == "__main__":
    test_docstrings()



#!/usr/bin/env python3

# Tests for ViZDoom enums and related methods.
# This test can be run as Python script or via PyTest

import vizdoom as vzd


def _test_enums(enum_name, func_name):
    print(f"Testing vzd.{enum_name} enum ...")

    game = vzd.DoomGame()
    game.set_window_visible(False)

    add_func = eval("game.add_" + func_name)
    get_func = eval("game.get_" + func_name + "s")
    set_func = eval("game.set_" + func_name + "s")
    clear_func = eval("game.clear_" + func_name + "s")

    all_values = [
        eval("vzd." + enum_name + "." + v)
        for v in dir(eval("vzd." + enum_name))
        if not v.startswith("__")
        and not v.startswith("_pybind11")
        and not v == "name"
        and not v == "value"
    ]

    all_values_names = [v.name for v in all_values]

    # set_X function test
    set_func(all_values)
    get_values_names = [v.name for v in get_func()]
    assert all_values_names == get_values_names

    # add_X function test
    clear_func()
    for i, v in enumerate(all_values):
        add_func(v)
        get_values_names = [v.name for v in get_func()]
        assert all_values_names[: i + 1] == get_values_names

    # Check if set function overwrites previous values
    set_func(all_values)
    get_values_names = [v.name for v in get_func()]
    assert all_values_names == get_values_names

    # Multiple add_X functions test
    for i, v in enumerate(all_values):
        add_func(v)
        get_values_names = [v.name for v in get_func()]
        assert all_values_names == get_values_names

    # Test duplicated values in set_X function
    set_func(all_values + all_values)
    get_values_names = [v.name for v in get_func()]
    assert all_values_names == get_values_names


def test_buttons():
    _test_enums("Button", "available_button")


def test_game_variables():
    _test_enums("GameVariable", "available_game_variable")


if __name__ == "__main__":
    test_buttons()
    test_game_variables()



#!/usr/bin/env python3
"""
Automated ViZDoom Type Stub Generator

This script uses pybind11-stubgen to generate stub file for ViZDoom, part of build target `generate_stubs`.

Workflow:
- using `pybind11-stubgen` for generation
- removing the `__all__ = ...` line
- adding `import numpy as np` after `import typing`
- optionally, `black` and `isort` for simple formatting
- adding a header comment
- replacing the `-> typing.Any` properties with actual return behaviour (`np.ndarray` or `typing.Optional[np.ndarray]`)
"""

import argparse
import os
import re
import shutil
import subprocess
import sys
import tempfile


vizdoom_stub_header = '''"""
ViZDoom Python Type Stubs

This file provides type information for static analysis and IDE support.
Auto-generated via pybind11-stubgen{formatted}.

For the official documentation, see: https://vizdoom.farama.org/
"""

'''


class ViZDoomStubGenerator:
    """Generator for ViZDoom type stubs with additional"""

    def __init__(
        self,
        output_file: str = "src/lib_python/vizdoom.pyi",
        verbose: bool = False,
        module: str = "vizdoom",
    ):
        self.output_file = output_file
        self.temp_dir = tempfile.mkdtemp()
        self.docstrings = {}
        self.enum_fixes = {}
        self.verbose = bool(verbose)
        self.module = module

    def generate_with_stubgen(self):
        """Generate basic type stubs using stubgen."""
        if self.verbose:
            print("Generating type stubs with stubgen...")

        # Add module directory to Python path if provided
        env: dict[str, str] = os.environ.copy()
        if os.path.isdir(self.module):
            print("module", self.module)
            module_path = self.module
            if self.verbose:
                print(f"Adding {module_path} to PYTHONPATH")
            current_path = env.get("PYTHONPATH", "")
            pathsep = os.pathsep
            paths = current_path.split(pathsep) if current_path else []
            if module_path not in paths:
                paths.insert(0, module_path)
                env["PYTHONPATH"] = pathsep.join(paths)

        # Run stubgen on the compiled vizdoom module
        cmd = ["pybind11-stubgen", "vizdoom", "-o", self.temp_dir]

        try:
            subprocess.run(cmd, capture_output=True, text=True, check=True, env=env)
            if not os.path.exists(os.path.join(self.temp_dir, "vizdoom.pyi")):
                self.temp_dir = os.path.join(self.temp_dir, "vizdoom")
            stub_file = os.path.join(self.temp_dir, "vizdoom.pyi")
            if os.path.exists(stub_file):
                with open(stub_file, "r", encoding="utf-8", errors="ignore") as f:
                    stub_content = []
                    for line in f:
                        if line.startswith("import typing"):
                            stub_content += (
                                "import typing\nfrom numpy.typing import NDArray\n"
                            )
                        elif not line.startswith("__all__ = "):
                            stub_content += line
                with open(stub_file, "w", encoding="utf-8", errors="ignore") as f:
                    f.writelines(stub_content)
            else:
                raise FileNotFoundError(
                    f"Stubgen did not create expected file: {stub_file}"
                )

        except subprocess.CalledProcessError as e:
            print(f"Stubgen failed: {e}")
            print(f"STDOUT: {e.stdout}")
            print(f"STDERR: {e.stderr}")
            raise

    def reformat_generated_stub(self):
        """Reformat generated type stubs with black and isort."""
        finished_formatting: list[str] = []
        try:
            for tool in ["black", "isort"]:
                if self.verbose:
                    print(f"Reformatting type stubs with {tool}...")
                cmd = [tool, os.path.join(self.temp_dir, "vizdoom.pyi")]
                subprocess.run(cmd, capture_output=True, text=True, check=True)
                finished_formatting.append(tool)
        except subprocess.CalledProcessError as e:
            print(f"{tool} failed: {e}")
            print(f"STDOUT: {e.stdout}")
            print(f"STDERR: {e.stderr}")
        return finished_formatting

    def load_generated_stub(self) -> str:
        """Reformat generated type stubs with black."""
        if self.verbose:
            print("Loading generated type stubs...")
        stub_file = os.path.join(self.temp_dir, "vizdoom.pyi")
        if os.path.exists(stub_file):
            with open(stub_file, "r") as f:
                content = f.read().removeprefix('"""\nViZDoom Python module.\n"""\n')
            # Remove exact value of __version__ from the stub to avoid misguiding the user
            content = re.sub(r"^(__version__: str).*$", r"\1", content, count=1, flags=re.MULTILINE)
            return content
        else:
            raise FileNotFoundError(
                f"Stubgen did not create expected file: {stub_file}"
            )

    def add_module_header(self, content: str, formatted_with: list[str]) -> str:
        """Add a module header."""
        if self.verbose:
            print("Adding module header...")
        if formatted_with:
            formatted = f" and formatted with {', '.join(formatted_with)}"
        else:
            formatted = ""
        return vizdoom_stub_header.format(formatted=formatted) + content

    def annotate_gamestate_properties(self, content: str) -> str:
        """Additional treatment for properties of GameState."""
        if self.verbose:
            print("Annotating properties of the GameState class...", end=" ")
        replacements = []
        match_game_state_properties = re.finditer(
            r"^(\s*def (screen|depth|audio|automap|labels|game)_"
            r"(?:buffer|variables)\(self\)\s*->\s*)(typing\.)?Any\s*(:.*)$",
            content,
            flags=re.MULTILINE,
        )
        for match_property in match_game_state_properties:
            if match_property.group(2) == "screen":
                return_type = "NDArray"
            else:
                return_type = match_property.group(3) + "Optional[NDArray]"
            actual_return_type = (
                match_property.group(1) + return_type + match_property.group(4)
            )
            replacements.append((match_property.group(0), actual_return_type))

        for old_line, new_line in replacements:
            content = content.replace(old_line, new_line, 1)

        if self.verbose:
            print(f"(patched {len(replacements)} lines)")

        return content

    def generate(self) -> str:
        """Generate complete type stubs with docstrings."""
        if self.verbose:
            print("Starting ViZDoom stub generation...")

        try:
            # Step 1: Generate basic stubs with stubgen
            self.generate_with_stubgen()

            # Step 2: Reformat stub file with black & isort
            formatted_with = self.reformat_generated_stub()

            # Step 3: Load stub filex
            stub_content = self.load_generated_stub()

            # Step 4: Add module header
            stub_content = self.add_module_header(stub_content, formatted_with)

            # Step 5: Additionally treatment for properties of GameState object
            stub_content = self.annotate_gamestate_properties(stub_content)

            # Step 5.5: The small patches
            stub_content = stub_content.replace(
                "set_config(self, config: typing.Any",
                "set_config(self, config: typing.Union[str, dict[str, typing.Any]]",
            )

            # Step 6: Write to output file
            output_dir = os.path.dirname(self.output_file)
            try:
                os.makedirs(output_dir, exist_ok=True)
            except OSError:
                pass  # maybe the folder is in use or something OS-dependent
            with open(self.output_file, "w") as f:
                f.write(stub_content)

            if self.verbose:
                print(
                    f"Generated {self.output_file} ({len(stub_content.splitlines())} lines)"
                )

            return stub_content

        except Exception as e:
            print(f"Stub generation failed: {e}")
            raise
        finally:
            # Cleanup temp directory
            shutil.rmtree(self.temp_dir, ignore_errors=True)


def main():
    """Main entry point."""
    parser = argparse.ArgumentParser(description="Generate ViZDoom type stubs")
    parser.add_argument(
        "-o",
        "--output",
        default="src/lib_python/vizdoom.pyi",
        help="Output file path (default: src/lib_python/vizdoom.pyi)",
    )
    parser.add_argument(
        "-m",
        "--module",
        default="vizdoom",
        help="Directory containing vizdoom module (default: vizdoom)",
    )
    parser.add_argument(
        "-p",
        "--patch",
        action="store_true",
        help="Patch ViZDoom library in current environment",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")

    args = parser.parse_args()

    generator = ViZDoomStubGenerator(args.output, args.verbose, args.module)
    try:
        import pybind11_stubgen

        assert pybind11_stubgen
        generator.generate()
        print("Successfully generated ViZDoom type stubs!")
        print(f"Output: {args.output}")
        if args.patch:
            import vizdoom

            vizdoom_loc = os.path.dirname(vizdoom.__file__)
            shutil.copyfile(
                args.output,
                os.path.join(vizdoom_loc, "vizdoom.pyi"),
                follow_symlinks=False,
            )
            with open(os.path.join(vizdoom_loc, "py.typed"), "w") as f:
                f.write("partial\n")
            print(f"Patched ViZDoom at: {vizdoom_loc}")
            print(f"Enjoy your type-hinted ViZDoom!")
        return 0
    except ImportError:
        print(
            f"Need to install pybind11-stubgen via:\npip install pybind11-stubgen\n"
        )
        print(
            "Optionally also install black and isort for formatting:\npip install black isort\n"
        )
    except Exception as e:
        print(f"Failed to generate stubs: {e}")
    return 1


if __name__ == "__main__":
    sys.exit(main())



"""
Implementation of the base Gymnasium environment for ViZDoom.

The first version was based on Gym interface by [Simon Hakenes](https://github.com/shakenes/vizdoomgym),
and developed by
[Arjun KG](https://github.com/arjun-kg),
[Benjamin Noah Beal](https://github.com/bebeal),
[Lawrence Francis](https://github.com/ldfrancis),
[Mark Towers](https://github.com/pseudo-rnd-thoughts).

Modified and updated by
[Muhammad Elnimr](https://github.com/melnimr),
[Hugo Huang](https://github.com/Trenza1ore),
[Marek Wydmuch](https://github.com/mwydmuch).

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import itertools
import warnings
from typing import Any, Optional

import gymnasium as gym
import numpy as np
import pygame
from gymnasium.utils import EzPickle

import vizdoom.vizdoom as vzd


# A fixed set of colors for each potential label
# for rendering an image.
# 256 is not nearly enough for all IDs, but we limit
# ourselves here to avoid hogging too much memory.
LABEL_COLORS = (
    np.random.default_rng(42).uniform(25, 256, size=(256, 3)).astype(np.uint8)
)

ASCII_CHARS = [chr(i) for i in range(0, 127)]


class VizdoomEnv(gym.Env, EzPickle):
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": vzd.DEFAULT_TICRATE,
    }

    def __init__(
        self,
        config_file: Optional[str] = None,
        frame_skip: int = 1,
        max_buttons_pressed: int = 0,
        render_mode: Optional[str] = None,
        treat_episode_timeout_as_truncation: bool = True,
        use_multi_binary_action_space: bool = True,
        **kwargs: Any,
    ):
        """
        Base class for Gymnasium interface for ViZDoom.
        Child classes are defined in gymnasium_env_defns.py,

        Arguments:
            config_file (Optional[str]): The path to the config file to load.
                                         Most settings should be set by this config file.
                                         If None, all the necessary config must be provided in **kwargs.
            frame_skip (int): The number of frames the will be advanced per action. 1 = take action on every frame. Default: 1.
            max_buttons_pressed (int): Defines the number of binary buttons that can be selected at once. Default: 1.
                                       Should be >= 0. If < 0 a RuntimeError is raised.
                                       If == 0, the binary action space becomes ``MultiBinary(len(num_binary_buttons))``
                                       or ``MultiDiscrete([2] * num_binary_buttons)`` (depending on ``use_multi_binary_action_space`` flag)
                                       and [0, ``num_binary_buttons``] number of binary buttons can be selected.
                                       If > 0, the binary action space becomes ``Discrete(n)``
                                       and ``n`` actions can be selected.
                                       ``n`` is equal to number of possible buttons combinations
                                       with the number of buttons pressed < ``max_buttons_pressed``.
            render_mode (Optional[str]): The render mode to use could be either "human" or "rgb_array"
            skill_level (Optional[int]): If specified, sets the skill level (difficulty) of the game (overrides config file).
                                         Valid values are 1 to 5, where 1 is the easiest and 5 is the hardest.
            map (Optional[str]): If specified, sets the map to start. Should be a valid map ID defined in the WAD file (overrides config file).
            treat_episode_timeout_as_truncation (bool): If True, the episode will be treated as truncated
                                                        when the internal episode timeout is reached.
                                                        This is compatibility option, ViZDoom versions <1.3.0 behave as if this was set to False.
                                                        Default: True.
            use_multi_binary_action_space (bool): If True, the ``MultiBinary(len(num_binary_buttons))`` action space
                                                  will be used for buttons binary buttons instead of ``MultiDiscrete([2] * len(num_binary_buttons))``.
                                                  This is compatibility option, ViZDoom versions <1.3.0 behave as if this was set to False.
                                                  Default: True.
            **kwargs: Additional config options to set in the DoomGame after loading the config file.

        This environment forces the game window to be hidden. Use :meth:`render` function to see the game.

        Observations are dictionaries with different number of entries, depending on if depth/label buffers were
        enabled in the config file (CHANNELS == 1 if GRAY8, else 3):

        - "screen": The screen image buffer (always available) as `np.ndarray` in shape ``(HEIGHT, WIDTH, CHANNELS)``.
        - "depth": The depth image as ``np.ndarray`` in shape ``(HEIGHT, WIDTH, 1)``, if enabled by the config file.
        - "labels": The label image buffer as `np.ndarray` in shape ``(HEIGHT, WIDTH, 1)``, if enabled by the config file.
          For info on labels, access `env.state.labels` variable.
        - "automap": The automap image buffer as ``np.ndarray`` in shape ``(HEIGHT, WIDTH, CHANNELS)``, if enabled by the config file.
        - "gamevariables": All game variables, in the order specified by the config file.

        Action space can be a single one of binary/continuous action space, or a ``Dict`` containing both:

        - "binary": Is ``Discrete(num_binary_buttons + 1)`` if :attr:`max_buttons_pressed` >= 1
          or ``MultiBinary(len(num_binary_buttons))`` if :attr:`max_buttons_pressed` == 0 and :attr:`use_multi_binary_action_space` is True,
          or ``MultiDiscrete([2] * num_binary_buttons)`` if :attr:`max_buttons_pressed` == 0 and :attr:`use_multi_binary_action_space` is False,
        - "continuous": Is ``Box(float32.min, float32.max, (num_delta_buttons,), float32)``.
        """
        EzPickle.__init__(
            self,
            config_file,
            frame_skip,
            max_buttons_pressed,
            render_mode,
            treat_episode_timeout_as_truncation,
            use_multi_binary_action_space,
            **kwargs,
        )
        self.frame_skip = frame_skip
        self.render_mode = render_mode
        self.treat_episode_timeout_as_truncation = treat_episode_timeout_as_truncation
        self.use_multi_binary_action_space = use_multi_binary_action_space

        # init game
        self.game = vzd.DoomGame()

        if config_file is not None:
            self.game.load_config(config_file)
        if kwargs is not None and len(kwargs) > 0:
            self.game.set_config(kwargs)
        if config_file is None and not kwargs:
            raise RuntimeError("Either config_file or kwargs must be provided.")

        if (
            kwargs is None or "window_visible" not in kwargs
        ):  # Gymnasium environments should not create a visible window by default, but allow users to override this in kwargs
            self.game.set_window_visible(False)

        if (
            kwargs is None or "audio_buffer_size" not in kwargs
        ):  # Gymnasium environments should have buffer size set to frame_skip for all buffers by default, but allow users to override this in kwargs
            self.game.set_audio_buffer_size(frame_skip)
        elif self.game.is_audio_buffer_enabled():
            warnings.warn(
                "audio_buffer_size is set in kwargs. Gymnasium wrapper sets this buffer size to frame_skip by default."
            )

        if (
            kwargs is None or "notifications_buffer_size" not in kwargs
        ):  # Gymnasium environments should have buffer size set to frame_skip for all buffers by default, but allow users to override this in kwargs
            self.game.set_notifications_buffer_size(frame_skip)
        elif self.game.is_notifications_buffer_enabled():
            warnings.warn(
                "notifications_buffer_size is set in kwargs. Gymnasium wrapper sets this buffer size to frame_skip by default."
            )

        screen_format = self.game.get_screen_format()
        if (
            screen_format != vzd.ScreenFormat.RGB24
            and screen_format != vzd.ScreenFormat.GRAY8
        ):
            if (
                kwargs is not None and "screen_format" in kwargs
            ):  # Only warn if user explicitly set screen_format in kwargs
                warnings.warn(
                    f"Detected screen format {screen_format.name} set in kwargs. "
                    f"Only RGB24 and GRAY8 are supported in the Gymnasium wrapper. Forcing RGB24."
                )
            self.game.set_screen_format(vzd.ScreenFormat.RGB24)

        self.state = None
        self.clock = None
        self.window_surface = None
        self.isopen = True
        self.channels = 3

        if screen_format == vzd.ScreenFormat.GRAY8:
            self.channels = 1

        self.depth = self.game.is_depth_buffer_enabled()
        self.labels = self.game.is_labels_buffer_enabled()
        self.automap = self.game.is_automap_buffer_enabled()
        self.audio = self.game.is_audio_buffer_enabled()
        self.notifications = self.game.is_notifications_buffer_enabled()

        # parse buttons defined by config file
        self.__parse_available_buttons()

        # check for valid max_buttons_pressed
        if max_buttons_pressed > self.num_binary_buttons > 0:
            warnings.warn(
                f"max_buttons_pressed={max_buttons_pressed} "
                f"> number of binary buttons defined={self.num_binary_buttons}. "
                f"Clipping max_buttons_pressed to {self.num_binary_buttons}."
            )
            max_buttons_pressed = self.num_binary_buttons
        elif max_buttons_pressed < 0:
            raise RuntimeError(
                f"max_buttons_pressed={max_buttons_pressed} < 0. Should be >= 0. "
            )

        # specify action space(s)
        self.max_buttons_pressed = max_buttons_pressed
        self.action_space = self.__get_action_space()

        # specify observation space(s)
        self.observation_space = self.__get_observation_space()

    def step(self, action: Any):
        assert self.action_space.contains(
            action
        ), f"{action!r} ({type(action)}) invalid"
        assert self.state is not None, "Call `reset` before using `step` method."

        env_action = self.__build_env_action(action)
        reward = self.game.make_action(env_action, self.frame_skip)
        self.state = self.game.get_state()
        terminated = self.game.is_episode_finished()
        truncated = (
            self.game.is_episode_timeout_reached()
            if self.treat_episode_timeout_as_truncation
            else False
        )
        if self.render_mode == "human":
            self.render()
        return self.__collect_observations(), reward, terminated, truncated, {}

    def __parse_binary_buttons(self, env_action: Any, agent_action: Any):
        if self.num_binary_buttons != 0:
            if self.num_delta_buttons != 0:
                agent_action = agent_action["binary"]

            if np.issubdtype(type(agent_action), np.integer):
                agent_action = self.button_map[agent_action]

            # binary actions offset by number of delta buttons
            env_action[self.num_delta_buttons :] = agent_action

    def __parse_delta_buttons(self, env_action: Any, agent_action: Any):
        if self.num_delta_buttons != 0:
            if self.num_binary_buttons != 0:
                agent_action = agent_action["continuous"]

            # delta buttons have a direct mapping since they're reorganized to be prior to any binary buttons
            env_action[0 : self.num_delta_buttons] = agent_action

    def __build_env_action(self, agent_action: Any):
        # encode users action as environment action
        env_action = np.array(
            [0 for _ in range(self.num_delta_buttons + self.num_binary_buttons)],
            dtype=np.float32,
        )
        self.__parse_delta_buttons(env_action, agent_action)
        self.__parse_binary_buttons(env_action, agent_action)
        return env_action

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        # Ensure DoomGame is always seeded with an unsigned int from gymnasium.Env's internal RNG
        game_seed = int(
            self.np_random.integers(0, np.iinfo(np.uint32).max + 1, dtype=np.uint32)
        )
        self.game.set_seed(game_seed)

        if self.game.is_running():
            self.game.new_episode()
        else:
            self.game.init()
        self.state = self.game.get_state()

        return self.__collect_observations(), {}

    def __collect_observations(self):
        observation = {}
        if self.state is not None:
            observation["screen"] = self.state.screen_buffer
            if self.channels == 1:
                observation["screen"] = self.state.screen_buffer[..., None]  # type: ignore
            if self.depth:
                observation["depth"] = self.state.depth_buffer[..., None]  # type: ignore
            if self.labels:
                observation["labels"] = self.state.labels_buffer[..., None]  # type: ignore
            if self.automap:
                observation["automap"] = self.state.automap_buffer
                if self.channels == 1:
                    observation["automap"] = self.state.automap_buffer[..., None]  # type: ignore
            if self.audio:
                observation["audio"] = self.state.audio_buffer
            if self.notifications:
                observation["notifications"] = self.state.notifications_buffer
            if self.num_game_variables > 0:
                observation["gamevariables"] = self.state.game_variables.astype(  # type: ignore
                    np.float32
                )
        else:
            # there is no state in the terminal step, so a zero observation is returned instead
            for space_key, space_item in self.observation_space.spaces.items():
                if isinstance(space_item, gym.spaces.Box):
                    observation[space_key] = np.zeros(
                        space_item.shape, dtype=space_item.dtype
                    )
                elif isinstance(space_item, gym.spaces.Text):
                    observation[space_key] = ""
                else:
                    warnings.warn(
                        f"Observation space of type {type(space_item)} not supported when there is no game state."
                    )

        return observation

    def __build_human_render_image(self):
        """Stack all available buffers into one for human consumption"""
        game_state = self.game.get_state()
        valid_buffers = game_state is not None

        if not valid_buffers:
            # Return a blank image
            num_enabled_buffers = 1 + self.depth + self.labels + self.automap
            img = np.zeros(
                (
                    self.game.get_screen_height(),
                    self.game.get_screen_width() * num_enabled_buffers,
                    3,
                ),
                dtype=np.uint8,
            )
            return img

        image_list = [game_state.screen_buffer]
        if self.channels == 1:
            image_list = [
                np.repeat(game_state.screen_buffer[..., None], repeats=3, axis=2)
            ]

        if self.depth:
            image_list.append(
                np.repeat(game_state.depth_buffer[..., None], repeats=3, axis=2)  # type: ignore
            )

        if self.labels:
            # Give each label a fixed color.
            # We need to connect each pixel in labels_buffer to the corresponding
            # id via `value``
            labels_rgb = np.zeros_like(image_list[0])
            labels_buffer = game_state.labels_buffer
            for label in game_state.labels:
                color = LABEL_COLORS[label.object_id % 256]
                labels_rgb[labels_buffer == label.value] = color
            image_list.append(labels_rgb)

        if self.automap:
            automap_buffer = game_state.automap_buffer
            if self.channels == 1:
                automap_buffer = np.repeat(automap_buffer[..., None], repeats=3, axis=2)  # type: ignore
            image_list.append(automap_buffer)

        return np.concatenate(image_list, axis=1)

    def render(self):
        if self.clock is None:
            self.clock = pygame.time.Clock()
        render_image = self.__build_human_render_image()
        if self.render_mode == "rgb_array":
            return render_image
        elif self.render_mode == "human":
            # Transpose image (pygame wants (width, height, channels), we have (height, width, channels))
            render_image = render_image.transpose(1, 0, 2)
            if self.window_surface is None:
                pygame.init()
                pygame.display.set_caption("ViZDoom")
                self.window_surface = pygame.display.set_mode(render_image.shape[:2])

            surf = pygame.surfarray.make_surface(render_image)
            self.window_surface.blit(surf, (0, 0))
            pygame.display.update()
            pygame.event.pump()
            self.clock.tick(self.metadata["render_fps"])
        else:
            return self.isopen

    def close(self):
        self.game.close()
        if self.window_surface:
            pygame.quit()
            self.isopen = False

    def __parse_available_buttons(self):
        """
        Parses the currently available game buttons,
        reorganizes all delta buttons to be prior to any binary buttons
        sets ``num_delta_buttons``, ``num_binary_buttons``
        """
        delta_buttons = []
        binary_buttons = []
        for button in self.game.get_available_buttons():
            if vzd.is_delta_button(button) and button not in delta_buttons:
                delta_buttons.append(button)
            else:
                binary_buttons.append(button)
        # force all delta buttons to be first before any binary buttons
        self.game.set_available_buttons(delta_buttons + binary_buttons)
        self.num_delta_buttons = len(delta_buttons)
        self.num_binary_buttons = len(binary_buttons)
        if delta_buttons == binary_buttons == 0:
            raise RuntimeError(
                "No game buttons defined. Must specify game buttons using `available_buttons` in the "
                "config file."
            )

    def __get_binary_action_space(self):
        """
        Return binary action space:
        ``Discrete(n)`` or ``MultiBinary(num_binary_buttons)`` or ``MultiDiscrete([2] * num_binary_buttons)``
        """
        if self.max_buttons_pressed == 0:
            if self.use_multi_binary_action_space:
                button_space = gym.spaces.MultiBinary(self.num_binary_buttons)
            else:
                button_space = gym.spaces.MultiDiscrete(
                    [
                        2,
                    ]
                    * self.num_binary_buttons
                )

        else:
            self.button_map = [
                np.array(list(action))
                for action in itertools.product((0, 1), repeat=self.num_binary_buttons)
                if (self.max_buttons_pressed >= sum(action) >= 0)
            ]
            button_space = gym.spaces.Discrete(len(self.button_map))
        return button_space

    def __get_continuous_action_space(self):
        """
        Returns continuous action space: Box(float32.min, float32.max, (num_delta_buttons,), float32)
        """
        return gym.spaces.Box(
            np.finfo(np.float32).min,
            np.finfo(np.float32).max,
            (self.num_delta_buttons,),
            dtype=np.float32,
        )

    def __get_action_space(self):
        """
        Returns action space:
            if both binary and delta buttons defined in the config file, action space will be:
              ``Dict("binary": Discrete|MultiBinary|MultiDiscrete, "continuous", Box)``
            else:
              action space will be only one of the following ``Discrete``|``MultiBinary``|``MultiDiscrete``|``Box``
        """
        if self.num_delta_buttons == 0:
            return self.__get_binary_action_space()
        elif self.num_binary_buttons == 0:
            return self.__get_continuous_action_space()
        else:
            return gym.spaces.Dict(
                {
                    "binary": self.__get_binary_action_space(),
                    "continuous": self.__get_continuous_action_space(),
                }
            )

    def __get_observation_space(self):
        """
        Returns observation space: Dict with Box entry for each activated buffer:
          "screen", "depth", "labels", "automap", "gamevariables"
        """
        spaces = {
            "screen": gym.spaces.Box(
                0,
                255,
                (
                    self.game.get_screen_height(),
                    self.game.get_screen_width(),
                    self.channels,
                ),
                dtype=np.uint8,
            )
        }

        if self.depth:
            spaces["depth"] = gym.spaces.Box(
                0,
                255,
                (self.game.get_screen_height(), self.game.get_screen_width(), 1),
                dtype=np.uint8,
            )

        if self.labels:
            spaces["labels"] = gym.spaces.Box(
                0,
                255,
                (self.game.get_screen_height(), self.game.get_screen_width(), 1),
                dtype=np.uint8,
            )

        if self.automap:
            spaces["automap"] = gym.spaces.Box(
                0,
                255,
                (
                    self.game.get_screen_height(),
                    self.game.get_screen_width(),
                    # "automap" buffer uses same number of channels
                    # as the main screen buffer,
                    self.channels,
                ),
                dtype=np.uint8,
            )
        if self.audio:
            spaces["audio"] = gym.spaces.Box(
                -32768,
                32767,
                (
                    int(
                        self.game.get_audio_sampling_rate() * 1 / 35 * self.frame_skip
                    ),  # rate / 35tics * frameskip
                    # 2 channels audio
                    2,
                ),
                dtype=np.int16,
            )
        if self.notifications:
            spaces["notifications"] = gym.spaces.Text(
                min_length=0, max_length=32768, charset=ASCII_CHARS
            )

        self.num_game_variables = self.game.get_available_game_variables_size()
        if self.num_game_variables > 0:
            spaces["gamevariables"] = gym.spaces.Box(
                np.finfo(np.float32).min,
                np.finfo(np.float32).max,
                (self.num_game_variables,),
                dtype=np.float32,
            )

        return gym.spaces.Dict(spaces)



#!/usr/bin/env python3

# Test correctness of labels buffer.
# This test can be run as Python script or via PyTest

import os
from random import choice

import numpy as np

import vizdoom as vzd


def check_label(labels_buffer, label):
    assert label.value > 1
    if (
        label.width > 4 and label.height > 4
    ):  # Sometimes very tiny objects may be obscured by level geometry or not rendered due to sprite size
        # Values decrease with the distance from the player,
        # check for >= since objects with higher values (closer ones) can obscure objects with lower values (further ones).
        detect_color = labels_buffer >= label.value
        detect_color[
            label.y : label.y + label.height, label.x : label.x + label.width
        ] = False  # Set values inside bounding box to False
        value_outside_box = np.any(
            detect_color >= label.value
        )  # Check if there is True outside the bounding box
        value_in_box = np.any(
            labels_buffer[
                label.y : label.y + label.height, label.x : label.x + label.width
            ]
            >= label.value
        )  # Check if there is True in the bounding box
        assert not value_outside_box and value_in_box


def test_labels_buffer():
    print("Testing labels buffer ...")
    game = vzd.DoomGame()
    game.load_config(os.path.join(vzd.scenarios_path, "deathmatch.cfg"))

    game.set_screen_resolution(vzd.ScreenResolution.RES_640X480)
    game.set_labels_buffer_enabled(True)
    game.set_render_hud(False)
    game.set_window_visible(False)
    # game.set_mode(vzd.Mode.SPECTATOR)  # For manual testing

    game.init()

    actions = [
        [True, False, False, False],
        [False, True, False, False],
        [False, False, True, False],
        [False, False, False, True],
    ]

    game.new_episode()
    state_count = 0
    seen_labels = 0
    seen_unique_objects = set()

    while not game.is_episode_finished():
        state = game.get_state()
        assert state is not None
        labels_buffer = state.labels_buffer
        game.make_action(choice(actions))

        state_count += 1
        seen_labels += len(state.labels)
        for label in state.labels:
            seen_unique_objects.add(label.object_name)
            check_label(labels_buffer, label)
    game.close()

    print(
        f"Seen {seen_labels} labels with {len(seen_unique_objects)} unique objects in {state_count} states."
    )


if __name__ == "__main__":
    test_labels_buffer()



#!/usr/bin/env python3

# Tests if basic game loop works correctly in different modes.
# This test can be run as Python script or via PyTest.

import vizdoom as vzd


def _test_basic_loop(mode, episodes=3, steps=10, frame_skip=1):
    game = vzd.DoomGame()
    game.set_mode(mode)
    game.set_window_visible(False)
    game.set_available_buttons(
        [vzd.Button.MOVE_LEFT, vzd.Button.MOVE_RIGHT, vzd.Button.ATTACK]
    )
    game.set_episode_start_time(35)
    game.init()

    # Just run a few steps to see if anything crashes
    for _ in range(episodes):
        game.new_episode()
        for _ in range(steps):
            if game.is_episode_finished():
                break

            if mode in {vzd.Mode.ASYNC_SPECTATOR, vzd.Mode.SPECTATOR}:
                game.advance_action(frame_skip)
            else:
                game.make_action([0] * game.get_available_buttons_size(), frame_skip)

    game.close()


def test_basic_loop():
    modes = [
        vzd.Mode.PLAYER,
        vzd.Mode.ASYNC_PLAYER,
        vzd.Mode.SPECTATOR,
        vzd.Mode.ASYNC_SPECTATOR,
    ]

    frame_skips = [1, 4]

    for mode in modes:
        for frame_skip in frame_skips:
            print(f"Testing mode: {mode}, frame_skip: {frame_skip}")
            _test_basic_loop(mode, frame_skip=frame_skip)


if __name__ == "__main__":
    test_basic_loop()
